@@ -10,6 +10,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdbool.h>
+#include <stdarg.h>
 
 #include <glib.h>
 #include <libsoup/soup.h>
@@ -30,9 +31,20 @@ struct sr_session_priv {
 	int submit_count;
 	sr_track_t *last_track;
 	int np_timer;
+
+	/* web-service */
+	char *api_url;
+	char *api_key;
+	char *api_secret;
+	char *session_key;
+	GQueue *love_queue;
+	GMutex *love_queue_mutex;
+	bool api_problems;
 };
 
 static void now_playing(sr_session_t *s, sr_track_t *t);
+static void ws_auth(sr_session_t *s);
+static void ws_love(sr_session_t *s);
 
 sr_session_t *
 sr_session_new(const char *url,
@@ -50,6 +62,8 @@ sr_session_new(const char *url,
 	priv->client_ver = g_strdup(client_ver);
 	priv->soup = soup_session_async_new();
 	priv->handshake_delay = 1;
+	priv->love_queue = g_queue_new();
+	priv->love_queue_mutex = g_mutex_new();
 	return s;
 }
 
@@ -63,6 +77,19 @@ sr_session_free(sr_session_t *s)
 
 	priv = s->priv;
 
+	while (!g_queue_is_empty(priv->love_queue)) {
+		sr_track_t *t;
+		t = g_queue_pop_head(priv->love_queue);
+		sr_track_free(t);
+	}
+	g_queue_free(priv->love_queue);
+	g_mutex_free(priv->love_queue_mutex);
+
+	g_free(priv->api_url);
+	g_free(priv->api_key);
+	g_free(priv->api_secret);
+	g_free(priv->session_key);
+
 	if (priv->np_timer)
 		g_source_remove(priv->np_timer);
 
@@ -120,6 +147,8 @@ sr_track_new(void)
 void
 sr_track_free(sr_track_t *t)
 {
+	if (!t)
+		return;
 	g_free(t->artist);
 	g_free(t->title);
 	g_free(t->album);
@@ -155,6 +184,15 @@ check_last(sr_session_t *s,
 	c = priv->last_track;
 	if (!c)
 		return;
+
+	if (c->rating == 'L' && priv->session_key) {
+		g_mutex_lock(priv->love_queue_mutex);
+		g_queue_push_tail(priv->love_queue, sr_track_dup(c));
+		g_mutex_unlock(priv->love_queue_mutex);
+		if (!priv->api_problems)
+			ws_love(s);
+	}
+
 	playtime = timestamp - c->timestamp;
 	/* did the last track played long enough? */
 	if ((playtime >= 240 || playtime >= c->length / 2) && c->length > 30)
@@ -460,6 +498,13 @@ sr_session_handshake(sr_session_t *s)
 
 	g_free(handshake_url);
 	g_free(auth);
+
+	if (priv->api_key) {
+		if (!priv->session_key)
+			ws_auth(s);
+		else
+			ws_love(s);
+	}
 }
 
 static void
@@ -472,10 +517,11 @@ drop_submitted(sr_session_t *s)
 	for (c = 0; c < priv->submit_count; c++) {
 		sr_track_t *t;
 		t = g_queue_pop_head(priv->queue);
-		sr_track_free(t);
-		if (g_queue_is_empty(priv->queue))
+		if (!t)
 			break;
+		sr_track_free(t);
 	}
+	priv->submit_count = 0;
 	g_mutex_unlock(priv->queue_mutex);
 
 	if (!g_queue_is_empty(priv->queue))
@@ -509,33 +555,41 @@ scrobble_cb(SoupSession *session,
 	    gpointer user_data)
 {
 	sr_session_t *s = user_data;
+	struct sr_session_priv *priv = s->priv;
 	const char *data, *end;
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL(message->status_code)) {
 		hard_failure(s);
-		return;
+		goto nok;
 	}
 
 	data = message->response_body->data;
 	end = strchr(data, '\n');
 	if (!end) /* really bad */
-		return;
+		goto nok;
 
-	if (strncmp(data, "OK", end - data) == 0)
+	if (strncmp(data, "OK", end - data) == 0) {
 		drop_submitted(user_data);
+		return;
+	}
 	else if (strncmp(data, "BADSESSION", end - data) == 0)
 		invalidate_session(s);
 	else
 		hard_failure(s);
+nok:
+	g_mutex_lock(priv->queue_mutex);
+	priv->submit_count = 0;
+	g_mutex_unlock(priv->queue_mutex);
+
 }
 
 #define ADD_FIELD(id, fmt, field) \
 	do { \
 		if ((field)) \
-		g_string_append_printf(data, "&" id "[%i]=%" fmt, i, (field)); \
+			g_string_append_printf(data, "&" id "[%i]=%" fmt, i, (field)); \
 		else \
-		g_string_append_printf(data, "&" id "[%i]=", i); \
-	} while(0);
+			g_string_append_printf(data, "&" id "[%i]=", i); \
+	} while (0);
 
 #define EXTRA_URI_ENCODE_CHARS "&+"
 
@@ -553,7 +607,7 @@ sr_session_submit(sr_session_t *s)
 		return;
 
 	g_mutex_lock(priv->queue_mutex);
-	if (g_queue_is_empty(priv->queue)) {
+	if (g_queue_is_empty(priv->queue) || priv->submit_count) {
 		g_mutex_unlock(priv->queue_mutex);
 		return;
 	}
@@ -640,7 +694,7 @@ now_playing_cb(SoupSession *session,
 		g_string_append_printf(data, "&" id "=%" fmt, (field)); \
 		else \
 		g_string_append_printf(data, "&" id "="); \
-	} while(0);
+	} while (0);
 
 static void
 now_playing(sr_session_t *s,
@@ -704,3 +758,233 @@ sr_session_set_proxy(sr_session_t *s, const char *url)
 		soup_uri = NULL;
 	g_object_set(priv->soup, "proxy-uri", soup_uri, NULL);
 }
+
+/* web-service */
+
+void
+sr_session_set_session_key(sr_session_t *s,
+			   const char *session_key)
+{
+	struct sr_session_priv *priv = s->priv;
+	priv->session_key = g_strdup(session_key);
+}
+
+void
+sr_session_set_api(sr_session_t *s,
+		   const char *api_url,
+		   const char *api_key,
+		   const char *api_secret)
+{
+	struct sr_session_priv *priv = s->priv;
+	priv->api_url = g_strdup(api_url);
+	priv->api_key = g_strdup(api_key);
+	priv->api_secret = g_strdup(api_secret);
+}
+
+struct ws_param {
+	const char *key;
+	const char *value;
+};
+
+static struct ws_param *
+param_new(const char *key, const char *value)
+{
+	struct ws_param *p = malloc(sizeof(*p));
+	p->key = key;
+	p->value = value;
+	return p;
+}
+
+static int
+param_compare(struct ws_param *a, struct ws_param *b)
+{
+	return strcmp(a->key, b->key);
+}
+
+static void ws_params(sr_session_t *s, char **params, ...) __attribute__((sentinel));
+
+static void
+ws_params(sr_session_t *s, char **params, ...)
+{
+	struct sr_session_priv *priv = s->priv;
+	GList *l = NULL;
+	va_list args;
+	GString *params_str, *tmp;
+	gchar *api_sig;
+
+	if (!params)
+		return;
+
+	params_str = g_string_sized_new(0x100);
+
+	va_start(args, params);
+	do {
+		const char *key, *value;
+		key = va_arg(args, char *);
+		if (!key)
+			break;
+		value = va_arg(args, char *);
+		if (!value)
+			break;
+		l = g_list_prepend(l, param_new(key, value));
+		g_string_append_printf(params_str, "&%s=%s", key, value);
+	} while (true);
+	va_end(args);
+
+	l = g_list_sort(l, (GCompareFunc) param_compare);
+
+	tmp = g_string_sized_new(0x100);
+
+	GList *c;
+	for (c = l; c; c = c->next) {
+		struct ws_param *p = c->data;
+		g_string_append_printf(tmp, "%s%s", p->key, p->value);
+	}
+
+	g_string_append(tmp, priv->api_secret);
+
+	api_sig = g_compute_checksum_for_string(G_CHECKSUM_MD5, tmp->str, -1);
+
+	g_string_free(tmp, TRUE);
+	g_list_foreach(l, (GFunc) free, NULL);
+
+	g_string_append_printf(params_str, "&api_sig=%s", api_sig);
+	g_free(api_sig);
+
+	*params = params_str->str;
+	g_string_free(params_str, FALSE);
+}
+
+static void
+ws_auth_cb(SoupSession *session,
+	   SoupMessage *message,
+	   void *user_data)
+{
+	sr_session_t *s = user_data;
+	struct sr_session_priv *priv = s->priv;
+	const char *data, *begin, *end;
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL(message->status_code))
+		return;
+
+	data = message->response_body->data;
+
+	begin = strstr(data, "<key>") + 5;
+	if (!begin) /* really bad */
+		return;
+	end = strstr(begin, "</key>");
+	if (!end) /* really bad */
+		return;
+	priv->session_key = g_strndup(begin, end - begin);
+	if (s->session_key_cb)
+		s->session_key_cb(s, priv->session_key);
+}
+
+static void
+ws_auth(sr_session_t *s)
+{
+	struct sr_session_priv *priv = s->priv;
+	gchar *auth, *tmp;
+	gchar *auth_url;
+	SoupMessage *message;
+
+	gchar *params;
+
+	tmp = g_strdup_printf("%s%s", priv->user, priv->hash_pwd);
+	auth = g_compute_checksum_for_string(G_CHECKSUM_MD5, tmp, -1);
+	g_free(tmp);
+
+	ws_params(s, &params,
+		  "api_key", priv->api_key,
+		  "authToken", auth,
+		  "method", "auth.getMobileSession",
+		  "username", priv->user,
+		  NULL);
+
+	auth_url = g_strdup_printf("%s?%s",
+				   priv->api_url,
+				   params);
+	g_free(params);
+
+	message = soup_message_new("GET", auth_url);
+	soup_session_queue_message(priv->soup,
+				   message,
+				   ws_auth_cb,
+				   s);
+
+	g_free(auth_url);
+	g_free(auth);
+}
+
+static void
+ws_love_cb(SoupSession *session,
+	   SoupMessage *message,
+	   void *user_data)
+{
+	sr_session_t *s = user_data;
+	struct sr_session_priv *priv = s->priv;
+	sr_track_t *t;
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL(message->status_code)) {
+		priv->api_problems = true;
+		return;
+	}
+
+	g_mutex_lock(priv->love_queue_mutex);
+	t = g_queue_pop_head(priv->love_queue);
+	g_mutex_unlock(priv->love_queue_mutex);
+	sr_track_free(t);
+
+	priv->api_problems = false;
+
+	if (!g_queue_is_empty(priv->love_queue))
+		/* still need to submit more */
+		ws_love(s);
+}
+
+static void
+ws_love(sr_session_t *s)
+{
+	struct sr_session_priv *priv = s->priv;
+	SoupMessage *message;
+	gchar *params;
+	sr_track_t *t;
+
+	g_mutex_lock(priv->love_queue_mutex);
+	t = g_queue_peek_head(priv->love_queue);
+	g_mutex_unlock(priv->love_queue_mutex);
+
+	if (!t)
+		return;
+
+	ws_params(s, &params,
+		  "method", "track.love",
+		  "api_key", priv->api_key,
+		  "sk", priv->session_key,
+		  "track", t->title,
+		  "artist", t->artist,
+		  NULL);
+
+	message = soup_message_new("POST", priv->api_url);
+	soup_message_set_request(message,
+				 "application/x-www-form-urlencoded",
+				 SOUP_MEMORY_TAKE,
+				 params,
+				 strlen(params));
+	soup_session_queue_message(priv->soup,
+				   message,
+				   ws_love_cb,
+				   s);
+}
+
+void
+sr_session_set_love(sr_session_t *s, int on)
+{
+	struct sr_session_priv *priv = s->priv;
+	sr_track_t *c;
+
+	c = priv->last_track;
+	if (!c)
+		return;
+	c->rating = on ? 'L' : '\0';
+}
